import os
from subprocess import Popen, PIPE

def ilen(it):
    return sum(1 for i in it)

def mercurial(path, ignore_set, **options):
    """Get statuses of a Mercurial repository."""
    ignore_untracked = options['ignore_untracked']

    process = Popen(('hg', 'st'), stdout=PIPE, cwd=path)
    output = process.stdout.read()
    lines = output.splitlines()
    untracked_count = ilen((line for line in lines if line.startswith('?')))
    untracked_only = untracked_count == len(lines)

    touched = bool(lines)
    if ignore_untracked and untracked_only:
        touched = False
    if touched:
        if untracked_only:
            status = 'untracked'
        else:
            status = 'uncommitted'
    else:
        status = 'OK'

    yield dict(
        touched=touched,
        path=path,
        status=status,
        output=output,
    )

    process = Popen(('hg', 'out'), stdout=PIPE, cwd=path)
    output = process.stdout.read()
    lines = []
    for line in output.splitlines():
        if not line.strip():
            continue
        if line.startswith('comparing with'):
            continue
        if line.startswith('searching for changes'):
            continue
        if line.startswith('no changes found'):
            continue
        lines.append(line)
    output='\n'.join(lines)+'\n'

    touched = bool(lines)
    if touched:
        yield dict(
            touched=touched,
            path=path,
            status='unpushed' if touched else 'OK',
            output=output,
        )

def git(path, ignore_set, **options):
    """Get statuses of a Git repository."""
    ignore_untracked = options['ignore_untracked']

    process = Popen(('git', 'status', '--porcelain'), stdout=PIPE, cwd=path)
    output = process.stdout.read()
    lines = output.splitlines()
    untracked_count = ilen((line for line in lines if line.startswith('?')))
    untracked_only = untracked_count == len(lines)

    touched = bool(lines)
    if ignore_untracked and untracked_only:
        touched = False
    if touched:
        if untracked_only:
            status = 'untracked'
        else:
            status = 'uncommitted'
    else:
        status = 'OK'

    yield dict(
        touched=touched,
        path=path,
        status=status,
        output=output,
    )

    # Related commands:
    # git log --branches --not --remotes --simplify-by-decoration --decorate --oneline
    # git branch -v
    # git for-each-ref --format='%(refname)' refs/heads
    process = Popen(('git', 'branch'), stdout=PIPE, cwd=path)
    branches = [br[2:] for br in process.stdout.read().splitlines()]
    for branch in branches:
        process = Popen(('git', 'log', branch, '--not', '--remotes',
            '--simplify-by-decoration',
            '--decorate', '--oneline', '--'), stdout=PIPE, cwd=path)
        output = process.stdout.read()
        touched = bool(output)
        yield dict(
            touched=touched,
            path='%s:%s' % (path, branch),
            status='unpushed' if touched else 'OK',
            output=output,
        )

def subversion(path, ignore_set, **options):
    """Get statuses of a Subversion repository."""
    ignore_untracked = options['ignore_untracked']

    process = Popen(('svn', 'st', '-v'), stdout=PIPE, cwd=path)
    output = process.stdout.read()
    lines = []
    for line in output.splitlines():
        if not line.strip():
            continue
        if line.startswith('Performing'):
            continue
        status = line[:8]
        filename = line[8:].split(None, 3)[-1]
        ignore_set.add(os.path.join(path, filename))
        if status.strip():
            lines.append(status + filename)
    output='\n'.join(lines)+'\n'

    untracked_count = ilen((line for line in lines if line.startswith('?')))
    untracked_only = untracked_count == len(lines)

    touched = bool(lines)
    if ignore_untracked and untracked_only:
        touched = False
    if touched:
        if untracked_only:
            status = 'untracked'
        else:
            status = 'uncommitted'
    else:
        status = 'OK'

    yield dict(
        touched=touched,
        path=path,
        status=status,
        output=output,
    )



"""The 'unpushed-notify' command-line tool."""

import os
import sys
from optparse import OptionParser

from . import scanner

USAGE = '''usage: %prog [options] path [path...]

  Checks the status of all Version Control repositories beneath the paths
  given on the command line.  Notify on OSD if some of them has changes.'''

def here(*args):
    return os.path.join(os.path.abspath(os.path.dirname(__file__)), *args)

def notify_linux(report):
    import os
    import re
    import getpass
    from subprocess import Popen, PIPE
    import pynotify
    w = Popen(('w', getpass.getuser()), stdout=PIPE).stdout.read().splitlines()[2:]
    displays = set()
    for entry in w:
        display = entry.split(None, 8)[2]
        displays.add(display)
    filtered = set()
    for display in displays:
        m = re.match(r'^(:\d+)\.\d+$', display)
        if m:
            root_display = m.group(1)
            if root_display not in displays:
                filtered.add(display)
        else:
            filtered.add(display)
    for display in filtered:
        os.environ['DISPLAY'] = display
        pynotify.init('unpushed-notify')
        icon = 'file://'+here('logo.png')
        n = pynotify.Notification('You have changes in working directory', report, icon)
        n.show()

def main():
    parser = OptionParser(usage=USAGE)
    parser.add_option('-l', '--locate', dest='use_locate', action='store_true',
                      help='use locate(1) to find repositories')
    parser.add_option('-w', '--walk', dest='use_walk', action='store_true',
                      help='manually walk file tree to find repositories')
    parser.add_option('-t', '--tracked', dest='ignore_untracked', action='store_true',
                      help='ignore untracked files in repositories')
    (options, args) = parser.parse_args()

    if not args:
        parser.print_help()
        exit(2)

    if options.use_locate and options.use_walk:
        sys.stderr.write('Error: you cannot specify both "-l" and "-w"\n')
        exit(2)

    if options.use_walk:
        find_repos = scanner.find_repos_by_walking
    else:
        find_repos = scanner.find_repos_with_locate

    repos = set()

    for path in args:
        path = os.path.abspath(path)
        if not os.path.isdir(path):
            sys.stderr.write('Error: not a directory: %s\n' % (path,))
            continue
        repos.update(find_repos(path))

    repos = sorted(repos)
    report = ''
    for status in scanner.scan_repos(repos, ignore_untracked=options.ignore_untracked):
        if status['touched']:
            report += '%s %s (%s)\n' % (status['path'], status['status'], status['vcs'])
    if report:
        if sys.platform.startswith('linux'):
            notify_linux(report)
        else:
            raise NotImplementedError('Notification is not implemented for this system')

if __name__ == '__main__':
    main()



"""The 'unpushed' command-line tool itself."""

import os
import sys
from optparse import OptionParser

from . import scanner

USAGE = '''usage: %prog [options] path [path...]

  Checks the status of all Version Control repositories beneath the paths
  given on the command line.  Any repositories with uncommitted changes
  are printed to standard out, along with the status of the files inside.'''

def main():
    parser = OptionParser(usage=USAGE)
    parser.add_option('-l', '--locate', dest='use_locate', action='store_true',
                      help='use locate(1) to find repositories')
    parser.add_option('-v', '--verbose', action='store_true',
                      help='for each repository print it\'s status')
    parser.add_option('-a', '--all', action='store_true', dest='print_all',
                      help='print every repository whether changed or not')
    parser.add_option('-w', '--walk', dest='use_walk', action='store_true',
                      help='manually walk file tree to find repositories')
    parser.add_option('-t', '--tracked', dest='ignore_untracked', action='store_true',
                      help='ignore untracked files in repositories')
    (options, args) = parser.parse_args()

    if not args:
        parser.print_help()
        exit(2)

    if options.use_locate and options.use_walk:
        sys.stderr.write('Error: you cannot specify both "-l" and "-w"\n')
        exit(2)

    if options.use_walk:
        find_repos = scanner.find_repos_by_walking
    else:
        find_repos = scanner.find_repos_with_locate

    repos = set()

    for path in args:
        path = os.path.abspath(path)
        if not os.path.isdir(path):
            sys.stderr.write('Error: not a directory: %s\n' % (path,))
            continue
        repos.update(find_repos(path))

    repos = sorted(repos)
    for status in scanner.scan_repos(repos, ignore_untracked=options.ignore_untracked):
        if status['touched'] or options.print_all:
            status_char = '*' if status['touched'] else ' '
            print status_char, status['path'], status['status'], '('+status['vcs']+')'
            if options.verbose:
                print status['output']

if __name__ == '__main__':
    main()



import os
from setuptools import setup, find_packages

def read(fname):
    return open(os.path.join(os.path.abspath(os.path.dirname(__file__)), fname)).read()

setup(
    name = 'unpushed',
    version = '1.1.0',
    description = u'Scan version control for uncommitted and unpushed changes',
    long_description = read('README.rst'),
    author = 'Dmitry Bashkatov',
    author_email = 'dbashkatov@gmail.com',
    url = 'http://github.com/nailgun/unpushed/',
    classifiers = [
        'Development Status :: 4 - Beta',
        'Environment :: Console', 'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Topic :: Software Development :: Version Control',
        'Topic :: Utilities',
    ],
    package_dir = {'': 'src'},
    packages = find_packages('src'),
    package_data = {'': ['*.png']},
    include_package_data = True,
    install_requires = [],
    entry_points = '[console_scripts]\n'
        'unpushed = unpushed.command:main\n'
        'unpushed-notify = unpushed.notify:main\n',
)
